def dataframe_has_required_columns(df, payload):
    columns = ['time', 'src_host', 'src_port', 'dst_host', 'dst_port', 'protocol']
    if payload:
        columns.append("payload")
    return set(df.columns) == set(columns)



import fastparquet
import glob
import msgpack
import os
import pickle
import pytest
import shutil

from intake_parquet.source import ParquetSource
import intake

intake.registry['parquet'] = ParquetSource  # because pytest defers import
here = os.path.dirname(__file__)
path = os.path.join(here, 'test.parq')
path2 = os.path.join(here, 'test2.parq')
pf = fastparquet.ParquetFile(path)
data = next(pf.iter_row_groups())
data2 = pf.to_pandas()


@pytest.mark.parametrize('url', [path,
                                 [os.path.join(path, 'part.0.parquet'),
                                  os.path.join(path, 'part.1.parquet')],
                                 path2])
@pytest.mark.parametrize('columns', [None,
                                     ['bhello', 'f', 'i32'],
                                     ['bhello']])
def test_source(url, columns):
    if columns:
        d = data[columns]
        d2 = data2[columns]
    else:
        d = data
        d2 = data2
    source = ParquetSource(url, columns=columns)
    source.discover()
    assert source.npartitions == 2
    assert source.shape == (None, len(d2.columns))
    assert source.dtype['bhello'] == 'object'
    part0 = source.read_partition(0)
    assert len(part0) == 1001
    assert part0.reset_index(drop=True).equals(d)
    it = source.read_chunked()
    assert next(it).reset_index(drop=True).equals(part0)
    assert next(it).reset_index(drop=True).equals(part0)
    with pytest.raises(StopIteration):
        next(it)
    with pytest.raises(IndexError):
        source.read_partition(5)
    parts = source.read()
    assert parts.reset_index(drop=True).equals(d2)


def test_discover_after_dask():
    source = ParquetSource(path)
    d = source.discover()
    df = source.to_dask()
    d2 = source.discover()
    assert isinstance(d['dtype'], dict)  # maybe should be zero-length df
    assert set(d['dtype']) == set(df.columns)  # order may not match for dict
    assert d2['dtype'] == d['dtype']
    assert d2['shape'] == d['shape']
    # dask index starts at 0 for each partition
    assert df.compute().reset_index(drop=True).equals(
        source.read().reset_index(drop=True))


def test_discover_serialize():
    source = ParquetSource(path)
    assert msgpack.packb(source.discover())


def test_pickle():
    source = ParquetSource(path)
    d = source.read()
    source2 = pickle.loads(pickle.dumps(source))
    d2 = source2.read()
    assert d.equals(d2)


def test_on_s3():
    pytest.importorskip('s3fs')
    s = ParquetSource('s3://MDtemp/gzip-nation.impala.parquet')
    s.read()
    assert s.shape == (None, 4)


def test_filter():
    import tempfile
    import pandas as pd
    import numpy as np
    d = str(tempfile.mkdtemp())
    try:
        df = pd.DataFrame({'a': np.random.randint(10, size=100),
                           'b': np.random.choice(['oi', 'hi'], size=100)})
        df.index.name = 'index'
        fastparquet.write(d, df, partition_on=['b'], file_scheme='hive',
                          write_index=True)
        s = ParquetSource(d, filters=[('b', '==', 'hi')])
        disc = s.discover()
        assert disc['npartitions'] == 1
        # TODO: this fails because the index appears as a column
        # assert disc['shape'] == df[df.b == 'hi'].shape
        out = s.read()
        assert 'oi' not in out.b
        assert df[df.b == 'hi'].a.equals(out.a)
    finally:
        import shutil
        shutil.rmtree(d)


def test_with_cache():
    import tempfile
    d = tempfile.mkdtemp()
    old = intake.config.conf['cache_dir']
    expected = fastparquet.ParquetFile(os.path.join(here, 'split')).to_pandas()
    try:
        intake.config.conf['cache_dir'] = d
        cat = intake.open_catalog(os.path.join(here, 'cache_cat.yaml'))
        s = cat.split()
        assert isinstance(s.cache[0], intake.source.cache.DirCache)
        outfiles = s.cache[0].load(s._urlpath, output=False)
        assert outfiles
        assert outfiles[0].startswith(s.cache_dirs[0])
        loc = s.cache[0]._path(s._urlpath)
        assert glob.glob(loc + '/*/*/*.parquet')
        assert s.read().reset_index(drop=True).equals(expected)
    finally:
        shutil.rmtree(d)
        intake.config.conf['cache_dir'] = old



from intake.source import base
from . import __version__


class ParquetSource(base.DataSource):
    """
    Source to load parquet datasets.

    Produces a dataframe.

    A parquet dataset may be a single file, a set of files in a single
    directory or a nested set of directories containing data-files.

    The implementation uses either fastparquet or pyarrow, select with the
    `engine=` kwarg.

    Keyword parameters accepted by this Source:

    - columns: list of str or None
        column names to load. If None, loads all

    - index: str or None
        column to make into the index of the dataframe. If None, may be
        inferred from the saved matadata in certain cases.

    - filters: list of tuples
        row-group level filtering; a tuple like ``('x', '>', 1)`` would mean
        that if a row-group has a maximum value less than 1 for the column
        ``x``, then it will be skipped. Row-level filtering is *not*
        performed.

    - engine: 'fastparquet' or 'pyarrow'
        Which backend to read with.


    - gather_statistics : bool or None (default).
        Gather the statistics for each dataset partition. By default,
        this will only be done if the _metadata file is available. Otherwise,
        statistics will only be gathered if True, because the footer of
        every file will be parsed (which is very slow on some systems).

    - see dd.read_parquet() for the other named parameters that can be passed through.
    """
    container = 'dataframe'
    name = 'parquet'
    version = __version__
    partition_access = True

    def __init__(self, urlpath, metadata=None,
                 storage_options=None, **parquet_kwargs):
        self._urlpath = urlpath
        self._storage_options = storage_options or {}
        self._kwargs = parquet_kwargs or {}
        self._df = None

        super(ParquetSource, self).__init__(metadata=metadata)

    def _get_schema(self):
        if self._df is None:
            self._df = self._to_dask()
        dtypes = {k: str(v) for k, v in self._df._meta.dtypes.items()}
        self._schema = base.Schema(datashape=None,
                                   dtype=dtypes,
                                   shape=(None, len(self._df.columns)),
                                   npartitions=self._df.npartitions,
                                   extra_metadata={})
        return self._schema

    def _get_partition(self, i):
        self._get_schema()
        return self._df.get_partition(i).compute()

    def read(self):
        """
        Create single pandas dataframe from the whole data-set
        """
        self._load_metadata()
        return self._df.compute()

    def to_spark(self):
        """Produce Spark DataFrame equivalent

        This will ignore all arguments except the urlpath, which will be
        directly interpreted by Spark. If you need to configure the storage,
        that must be done on the spark side.

        This method requires intake-spark. See its documentation for how to
        set up a spark Session.
        """
        from intake_spark.base import SparkHolder
        args = [
            ['read'],
            ['parquet', [self._urlpath]]
        ]
        sh = SparkHolder(True, args, {})
        return sh.setup()

    def to_dask(self):
        self._load_metadata()
        return self._df

    def _to_dask(self):
        """
        Create a lazy dask-dataframe from the parquet data
        """
        import dask.dataframe as dd
        urlpath = self._get_cache(self._urlpath)[0]
        self._df = dd.read_parquet(urlpath,
                                   storage_options=self._storage_options, **self._kwargs)
        self._load_metadata()
        return self._df

    def _close(self):
        self._df = None
